'''
Built-in, globally-available admin actions.
'''

from django.contrib import messages
# from django.contrib.admin import helpers
from django.contrib.admin.decorators import action
from django.contrib.admin.utils import model_ngettext


# from django.core.exceptions import PermissionDenied


@action(
    permissions=['delete'],
    description='Delete selected %(verbose_name_plural)s',
)
def delete_selected(modeladmin, request, queryset):

    # opts = modeladmin.model._meta
    # app_label = opts.app_label

    # Populate deletable_objects, a data structure of all related objects that
    # will also be deleted.
    # (
    #     deletable_objects,
    #     model_count,
    #     perms_needed,
    #     protected,
    # ) = modeladmin.get_deleted_objects(queryset, request)

    # The user has already confirmed the deletion.
    # Do the deletion and return None to display the change list view again.
    # if request.POST.get('post') and not protected:
    # if perms_needed:
    #     raise PermissionDenied
    n = queryset.count()
    for obj in queryset:
        obj_display = str(obj)
        modeladmin.log_deletion(request, obj, obj_display)

    modeladmin.delete_queryset(request, queryset)

    modeladmin.message_user(
        request,
        f'Successfully deleted {n} {model_ngettext(modeladmin.opts, n)}.',
        messages.SUCCESS,
    )

    return None

    # objects_name = model_ngettext(queryset)

    # if perms_needed or protected:
    #     title = _('Cannot delete %(name)s') % {'name': objects_name}
    # else:
    #     title = _('Are you sure?')

    # context = {
    #     **modeladmin.admin_site.each_context(request),
    #     'title': title,
    #     'objects_name': str(objects_name),
    #     'deletable_objects': [deletable_objects],
    #     'model_count': dict(model_count).items(),
    #     'queryset': queryset,
    #     'perms_lacking': perms_needed,
    #     'protected': protected,
    #     'opts': opts,
    #     'action_checkbox_name': helpers.ACTION_CHECKBOX_NAME,
    #     'media': modeladmin.media,
    # }

    # request.current_app = modeladmin.admin_site.name

    # # Display the confirmation page
    # return TemplateResponse(
    #     request,
    #     modeladmin.delete_selected_confirmation_template
    #     or [
    #         'admin/%s/%s/delete_selected_confirmation.html'
    #         % (app_label, opts.model_name),
    #         'admin/%s/delete_selected_confirmation.html' % app_label,
    #         'admin/delete_selected_confirmation.html',
    #     ],
    #     context,
    # )



from .brace import display_value, get_remote_url, update_field
from .exception import INVALID_INPUT, E, ErrorResponse
from .request import get_data


__all__ = [
    'E',
    'ErrorResponse',
    'INVALID_INPUT',

    'get_data',

    'update_field',
    'display_value',
    'get_remote_url'
]



from json import JSONDecodeError, loads

from django.http import HttpRequest, QueryDict


def json_loader(body: bytes) -> dict:
    try:
        data = loads(body)
        if isinstance(data, dict):
            return data
    except JSONDecodeError:
        pass

    return {}


def get_data(request: HttpRequest) -> QueryDict:
    data: QueryDict = getattr(request, request.method, QueryDict()).copy()

    if request.content_type == 'application/json':
        data.update(json_loader(request.body))

    return data



import logging
import re
from collections.abc import Iterable
from functools import update_wrapper
from typing import Any, Callable

from black_knight.admin import ModelAdmin, actions
from black_knight.admin.models import GroupAdmin, UserAdmin
from black_knight.admin.utils import E, get_data
from django.apps import apps
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import AlreadyRegistered
from django.contrib.auth import authenticate
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase
from django.http import HttpRequest, JsonResponse
from django.middleware.csrf import get_token
from django.shortcuts import render
from django.urls import reverse
from django.utils.text import capfirst
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_protect


logger = logging.getLogger('black_knight.AdminSite')


INVALID_LOGIN_DATA = E('Invalid Login Data!')


class AdminSite(admin.AdminSite):
    template = 'black-knight.html'
    default_avatar = settings.STATIC_URL + 'black_knight/default_avatar.jpg'
    user_avatar: str | Iterable[str] | Callable[['AdminSite', Any], str] = None
    registered_apps = {}

    def __init__(self, name='black_knight'):
        super().__init__(name)
        self._actions = {'delete_selected': actions.delete_selected}
        self._global_actions = self._actions.copy()

    def register(self, models, admin_class=None, **options):
        # check if models is a single model not an iterable
        if isinstance(models, ModelBase):
            models = [models]

        if admin_class is None:
            admin_class = ModelAdmin

        if not issubclass(admin_class, ModelAdmin):
            from django.contrib.auth.models import Group, User

            for model in models:
                # register a custom model admin for
                # default group and user model
                if model == Group:
                    admin_class = GroupAdmin

                if model == User:
                    admin_class = UserAdmin

        if not issubclass(admin_class, ModelAdmin):
            raise ValueError((
                'admin class that you register your model with, '
                'most be a sub class of black_knight.admin.ModelAdmin'
            ))

        for model in models:
            if model._meta.abstract:
                raise ImproperlyConfigured(
                    'The model %s is abstract, so it cannot be registered with admin.'
                    % model.__name__
                )

            if model in self._registry:
                registered_admin = str(self._registry[model])
                msg = 'The model %s is already registered ' % model.__name__

                if registered_admin.endswith('.ModelAdmin'):
                    # Most likely registered without a ModelAdmin subclass.
                    msg += 'in app %r.' % re.sub(
                        r'\.ModelAdmin$', '', registered_admin
                    )
                else:
                    msg += 'with %r.' % registered_admin

                raise AlreadyRegistered(msg)

            if not model._meta.swapped:
                app_label = model._meta.app_label
                model_name = model._meta.model_name

                if not app_label in self.registered_apps:
                    self.registered_apps[app_label] = {}

                self.registered_apps[app_label][model_name] = (
                    model,
                    admin_class(model, self)
                )

                self._registry[model] = admin_class(model, self)

    def admin_view(self, view, json, cacheable=False):

        def inner(request: HttpRequest, *args, **kwargs):
            if not self.has_permission(request):
                if json is False:
                    from django.contrib.auth.views import redirect_to_login

                    return redirect_to_login(
                        request.get_full_path(),
                        reverse('black_knight:login', current_app=self.name),
                    )

                return JsonResponse({
                    'error': 'Admin Login Required!',
                    'login_required': True
                }, status=401)

            return view(request, *args, **kwargs)

        if not cacheable:
            inner = never_cache(inner)

        if not getattr(view, 'csrf_exempt', False):
            inner = csrf_protect(inner)

        return update_wrapper(inner, view)

    def url_wrap(self, view, json=True, cacheable=False):
        def wrapper(*args, **kwargs):
            return self.admin_view(view, json, cacheable)(*args, **kwargs)

        return update_wrapper(wrapper, view)

    def get_api_urls(self):
        from django.urls import include, path

        api_urls = [
            path('user/', self.url_wrap(self.api_user), name='user'),
            path('index/', self.url_wrap(self.api_index), name='index'),
            path('log/', self.url_wrap(self.api_log), name='log'),
            path('login/', self.api_login, name='login'),
            path('logout/', self.url_wrap(self.api_logout), name='logout'),
        ]

        def get_model_url(item):
            model, model_admin = item
            app_label = model._meta.app_label
            model_name = model._meta.model_name

            return path(f'{app_label}/{model_name}/', include(model_admin.api_urls))

        api_urls += list(map(get_model_url, self._registry.items()))

        return api_urls

    def get_urls(self):
        from django.urls import include, path, re_path

        api_urls = self.get_api_urls()
        app_view = self.url_wrap(self.index, json=False)

        urlpatterns = [
            path('', app_view, name='index'),
            path('api/', include((api_urls, self.name), namespace='api')),
            path('login/', self.index, name='login'),
        ]

        def get_model_url(model):
            app_label = model._meta.app_label
            model_name = model._meta.model_name

            return re_path(rf'^{app_label}/{model_name}/(.*)', app_view)

        urlpatterns += list(map(get_model_url, self._registry.keys()))

        return urlpatterns

    @property
    def urls(self):
        return self.get_urls(), 'black_knight', self.name

    @property
    def base_url(self) -> str:
        return reverse('black_knight:index', current_app=self.name)

    def _build_app_dict(self, request, label=None):
        '''
        Build the app dictionary. The optional `label` parameter filters models
        of a specific app.
        '''
        app_dict: dict[str, Any] = {}

        if label:
            models = {
                m: m_a
                for m, m_a in self._registry.items()
                if m._meta.app_label == label
            }
        else:
            models = self._registry

        for model, model_admin in models.items():
            app_label = model._meta.app_label

            if not model_admin.has_module_permission(request):
                continue

            perms = model_admin.get_model_perms(request)

            if True not in perms.values():
                continue

            # info = (app_label, model._meta.model_name)

            model_dict = {
                'name': model._meta.model_name,
                'plural_name': capfirst(model._meta.verbose_name_plural),
                'object_name': model._meta.object_name,
                'icon': model_admin.icon,
                'perms': perms,
            }

            if app_label in app_dict:
                app_dict[app_label]['models'].append(model_dict)
            else:
                app_dict[app_label] = {
                    'name': apps.get_app_config(app_label).verbose_name,
                    'app_label': app_label,
                    'models': [model_dict],
                }

        return app_dict

    def get_user_avatar(self, user: Any) -> str:
        try:
            img = None

            if isinstance(self.user_avatar, str):
                return self.user_avatar

            if isinstance(self.user_avatar, Iterable):
                img = user

                for attr in self.user_avatar:
                    img = getattr(img, attr)

            elif callable(self.user_avatar):
                img = self.user_avatar(self, user)

            # not None and not str
            if img is not None and not isinstance(img, str):
                raise TypeError(f'unsupported image type {type(img)}')

            return img or self.default_avatar

        except Exception as e:
            logger.exception(e)
            logger.warning('an error happened while getting user_avatar')

        return self.default_avatar

    def index(self, request: HttpRequest, *args, **kwargs):
        context = {'base_url': self.base_url}
        return render(request, self.template, context)

    def api_user(self, request: HttpRequest):
        user = request.user

        response = {
            'username': user.username,
            'first_name': user.first_name,
            'last_name': user.last_name,
            'email': user.email,
            'avatar': self.get_user_avatar(user)
        }

        return JsonResponse(response)

    def api_index(self, request: HttpRequest):
        app_list = self.get_app_list(request)

        response = {
            'apps': app_list
        }

        return JsonResponse(response)

    def api_log(self, request: HttpRequest):
        try:
            from django.contrib.admin.models import LogEntry

            def get_date(datetime):
                return datetime.strftime('%Y-%m-%d %H:%M:%S')

            def GL(log: LogEntry):
                return {
                    'user': str(log.user),
                    'flag': log.action_flag,
                    'time': get_date(log.action_time),
                    'message': log.get_change_message(),
                    'repr': log.object_repr,
                    'url': None,
                    'content_type': log.content_type.name
                }

            logs = list(map(GL, LogEntry.objects.all()[:3]))

            return JsonResponse({'logs': logs})
        except E as e:
            return e.response

    def api_logout(self, request: HttpRequest):
        auth_logout(request)
        return JsonResponse({'message': 'Your Logout was Successful'})

    def api_login(self, request: HttpRequest):
        try:
            data = get_data(request)
            username = data.get('username')
            password = data.get('password')

            if username is None or password is None:
                raise INVALID_LOGIN_DATA

            user = authenticate(
                request=request,
                username=username,
                password=password
            )

            if user is None:
                raise INVALID_LOGIN_DATA

            if not user.is_active:
                raise E('This User is Inactive')

            if not user.is_staff:
                raise E('This User is not a Staff')

            auth_login(request, user)
            get_token(request)

            return JsonResponse({'message': 'Your Login was Successful!'})
        except E as e:
            return e.response



from django.db.models import fields
from django.db.models.fields import files, related


class BaseField:

    def base_info(self, **kwargs):
        infos = {
            'required': not self.blank,
            'help_text': self.help_text,
            'initial': self.get_default(),
            'label': self.verbose_name,
            'name': self.name,
        }

        if self.choices is not None:
            infos['choices'] = self.get_choices(include_blank=False)

        infos.update(kwargs)

        return infos


class CharField(BaseField, fields.CharField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'char',
            'max_length': self.max_length
        })


class TextField(BaseField, fields.TextField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'text',
        })


class BooleanField(BaseField, fields.BooleanField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'boolean',
        })


class DateField(BaseField, fields.DateField):

    def get_default(self):
        value = super().get_default()
        return value.date().isoformat()

    @property
    def info(self):
        return super().base_info(**{
            'type': 'date',
        })


class DateTimeField(BaseField, fields.DateTimeField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'datetime',
        })


class PositiveIntegerField(BaseField, fields.PositiveIntegerField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'int',
            'min': 0
        })


class PositiveBigIntegerField(BaseField, fields.PositiveBigIntegerField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'int',
            'min': 0
        })


class ImageField(BaseField, files.ImageField):

    @property
    def info(self):
        return super().base_info(**{
            'type': 'image'
        })


class ForeignKey(BaseField, related.ForeignKey):

    def clean(self, value, model_instance):
        value = super().clean(value, model_instance)

        manager = self.remote_field.model._default_manager
        qs = manager.complex_filter(self.get_limit_choices_to())

        if isinstance(value, qs.model):
            return value

        value = qs.get(**{self.remote_field.field_name: value})

        return value

    @property
    def info(self):
        return super().base_info(**{
            'type': 'foreign_key',
            'choices': self.get_choices(include_blank=False)
        })



from black_knight.admin.utils import INVALID_INPUT, ErrorResponse
from black_knight.admin.utils import display_value, get_data, update_field
from django.contrib import admin
from django.contrib.admin.utils import flatten_fieldsets, label_for_field
from django.contrib.admin.utils import lookup_field
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.core.paginator import InvalidPage
from django.db.models.fields.related import ForeignObjectRel, ManyToManyRel
from django.db.models.fields.related import OneToOneField
from django.http import HttpRequest, JsonResponse, QueryDict
from django.middleware.csrf import get_token
from django.utils.decorators import method_decorator
from django.views.decorators.http import require_GET, require_POST


require_GET_m = method_decorator(require_GET)
require_POST_m = method_decorator(require_POST)


class ModelAdmin(admin.ModelAdmin):
    icon: str | None = None

    def get_api_urls(self):
        from django.urls import path, re_path

        wrap = self.admin_site.url_wrap

        return [
            path('brace-result/', wrap(self.brace_result)),
            path('brace-info/', wrap(self.brace_info)),
            path('brace-actions/', wrap(self.brace_actions)),

            re_path('brace-form/(add|change)/', wrap(self.brace_form)),
            re_path('brace-form-submit/(add|change)/',
                    wrap(self.brace_form_submit))
        ]

    @property
    def api_urls(self):
        return self.get_api_urls()

    def get_braceresult_instance(self, request):
        '''
        Return a `BraceResult` instance based on `request`. 
        May raise `IncorrectLookupParameters`.
        '''
        from black_knight.admin.views.main import BraceResult

        return BraceResult(request, self)

    def brace_result(self, request: HttpRequest):
        '''display list of instances in the brace'''
        try:
            brace_result = self.get_braceresult_instance(request)

            return JsonResponse(brace_result.response)
        except InvalidPage:
            return ErrorResponse('Invalid Page', 400)

    def get_orders(self):
        meta = self.model._meta
        orders = []

        for field in meta.fields:
            if field.remote_field:
                for remote_field in field.related_model._meta.fields:
                    if not remote_field.remote_field:
                        orders.append(f'{field.name}__{remote_field.name}')
            else:
                orders.append(field.name)

        return orders

    @require_GET_m
    def brace_info(self, request: HttpRequest):
        '''Brace Info'''
        get_token(request)
        list_display = self.get_list_display(request)
        root_queryset = self.get_queryset(request)
        actions = self.get_action_choices(request, [])

        # func = self.get_actions(request)['delete_selected'][0]

        response = {
            'preserve_filters': self.preserve_filters,
            'show_search': bool(self.get_search_fields(request)),
            'search_help_text': self.search_help_text,
            'full_result_count': None,
            'empty_value_display': self.get_empty_value_display(),
        }

        # actions
        actions = [{'name': a[0], 'description': a[1]} for a in actions]
        response['actions'] = actions or None
        response['orders'] = self.get_orders()

        # headers
        def get_label(field):
            label = label_for_field(
                field,
                self.model,
                self
            )
            return label.strip()

        response['headers'] = list(map(get_label, list_display))

        if self.show_full_result_count:
            response['full_result_count'] = root_queryset.count()

        return JsonResponse(response)

    @require_POST_m
    def brace_actions(self, request: HttpRequest):
        data = get_data(request)
        items = data.get('items', [])
        action = self.get_actions(request).get(data.get('action'))

        if not action or not items:
            return INVALID_INPUT
        if items != 'all' and not isinstance(items, list):
            return INVALID_INPUT

        queryset = self.model._default_manager.get_queryset()
        if isinstance(items, list):
            queryset = queryset.filter(pk__in=items)

        if not queryset:
            return ErrorResponse('Items not found', 404)

        # action = (func, name, desc)
        action[0](self, request, queryset)

        return JsonResponse({'ok': 'action executed successfully'})

    def brace_form(self, request: HttpRequest, form_type: str):
        fieldsets = []
        base_fieldsets = self.get_fieldsets(request)
        readonly_fields = self.get_readonly_fields(request)
        meta = self.model._meta
        instance = None
        response = {}

        # default type is add
        if form_type == 'change':
            pk = get_data(request).get('pk')
            instance = self.get_object(request, pk)
            if instance is None:
                return INVALID_INPUT

        def get_field(field_name):

            if field_name in readonly_fields:
                return {
                    'type': 'readonly',
                    'name': field_name,
                }

            field = meta.get_field(field_name)
            # field.clean(...)
            # field.save_form_data(obj, data)
            info = getattr(field, 'info', {
                'type': 'unknown',
                'name': field.name
            })

            return info

        def get_field_value(field_dict: dict):
            field, _, value = lookup_field(field_dict['name'], instance, self)

            # if field.remote_field and value:
            #     if isinstance(field.remote_field, ManyToManyRel):
            #         field_dict['value'] = ', '.join(map(str, value.all()))
            #         return field_dict
            #     elif isinstance(field.remote_field, (ForeignObjectRel, OneToOneField)):
            #         field_dict['value'] = get_remote_url(field, value)
            #         return field_dict

            field_dict['value'] = display_value(field, value)
            return field_dict

        for fieldset in base_fieldsets:
            fields = map(get_field, fieldset[1]['fields'])

            if instance:
                fields = map(get_field_value, fields)

            fieldsets.append({
                'name': fieldset[0],
                'description': fieldset[1].get('description'),
                'fields': list(fields),
            })

        response['fieldsets'] = fieldsets
        response['label'] = str(instance) if instance else None

        return JsonResponse(response)

    @require_POST_m
    def brace_form_submit(self, request: HttpRequest, form_type: str):
        data = request.POST.copy()
        data.update(request.FILES)
        errors = {}
        change = form_type == 'change'

        if change:
            instance = self.get_object(request, data.get('pk'))
            if instance is None:
                return INVALID_INPUT
        else:
            instance = self.model()

        fields = flatten_fieldsets(self.get_fieldsets(request, instance))

        for name in fields:
            try:
                update_field(name, instance, data, change)
            except ValidationError as e:
                # TODO: make a better error system
                errors[name] = getattr(
                    e, 'message',
                    getattr(e, 'messages', ['Error'])[0]
                )

        if errors:
            return JsonResponse({
                'message': 'Value Errors',
                'fields': errors,
                'code': 400
            }, status=400)
        else:
            instance.save()

        # TODO: better messages
        return JsonResponse({'message': 'success', 'pk': instance.pk})



from django.http import JsonResponse


def ErrorResponse(message: str, code: int) -> JsonResponse:
    context = {'message': str(message), 'code': code}
    return JsonResponse(context, status=code)


# default error massage
DEM = 'Error while processing the Request'


class E(Exception):
    message: str
    code: int

    def __init__(self, message: str = DEM, code: int = 400):
        self.message = str(message)
        self.code = int(code)

    def __str__(self):
        return self.message

    def __index__(self):
        return self.code

    @property
    def response(self):
        context = {'message': self.message, 'code': self.code}
        return JsonResponse(context, status=self.code)


INVALID_INPUT = ErrorResponse('Invalid Input', 400)



from black_knight.admin import ModelAdmin
from django.conf import settings


class GroupAdmin(ModelAdmin):
    icon = settings.STATIC_URL + 'favicon.ico'


class UserAdmin(ModelAdmin):
    list_display = 'username', 'email', 'first_name', 'last_name', 'is_staff'
    list_filter = 'is_staff', 'is_superuser', 'is_active', 'groups'
    search_fields = 'username', 'first_name', 'last_name', 'email'
    ordering = 'username',
    filter_horizontal = (
        'groups',
        'user_permissions',
    )



import datetime
import decimal

from django.db import models
from django.urls import NoReverseMatch, reverse
from django.utils import formats, timezone


def get_remote_url(field, value):
    meta = field.remote_field.model._meta
    app_label = meta.app_label
    model_name = meta.model_name

    try:
        url = reverse('black_knight:index', current_app='black_knight')
        url += f'{app_label}/{model_name}/change/{value.pk}/'
        return 'link', url
    except NoReverseMatch:
        str(value)


def display_value(field, value):

    if field:
        if getattr(field, 'flatchoices', None):
            return dict(field.flatchoices).get(value)

        if isinstance(field, models.ImageField):
            return 'image', value.url if value else None

        if isinstance(field, models.ForeignKey):
            return 'foreign_key', value.pk, str(value)

    if value is None:
        return None

    if value == '':
        return None

    if isinstance(value, bool):
        return value

    if isinstance(value, datetime.datetime):
        return 'datetime', value.isoformat()

    if isinstance(value, datetime.date):
        return 'date', value.isoformat()
        # return formats.localize(timezone.template_localtime(value))

    if isinstance(value, datetime.time):
        # TODO: return an iso format version of it
        return formats.localize(value)

    if isinstance(value, (int, decimal.Decimal, float)):
        return value
        # return formats.number_format(value)

    if isinstance(value, (list, tuple)):
        return ', '.join(str(v) for v in value)

    # print(type(field))

    return str(value)


def update_field(name, instance, data, change):
    meta = instance._meta
    f_name = 'F_' + name

    if change and not f_name in data:
        return

    field = meta.get_field(name)
    initial = field.get_default()
    value = data.get(f_name, initial)
    # value = initial if field.disabled else value
    value = field.clean(value, instance)
    field.save_form_data(instance, value)
